import logging
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Dict, List, Tuple


# ported over from
# https://golang.org/src/debug/gosym/pclntab.go


SYMS_ENCODING = "utf-8"


logger = logging.getLogger(__name__)


@dataclass
class Sym:
    value: int = 0
    type: int = 0
    name: str = ""
    go_type: int = 0
    func: Optional["Func"] = field(default=None, repr=False)


@dataclass
class Func(Sym):
    sym: Sym = field(default_factory=Sym)
    entry: int = 0
    end: int = 0
    params: List[Sym] = field(default_factory=list)
    locals: List[Sym] = field(default_factory=list)
    frame_size: int = 0
    line_table: Optional["LineTable"] = field(default=None, repr=False)
    obj: Optional["Obj"] = None


@dataclass
class Obj:
    funcs: List[Func] = field(default_factory=list, repr=False)
    paths: Optional[List[Sym]] = None


class ByteOrder(Enum):
    LITTLE_ENDIAN = "little"
    BIG_ENDIAN = "big"

    def from_bytes(self, b: memoryview) -> int:
        return int.from_bytes(b, self.value)

    def u32(self, b: memoryview) -> int:
        return self.from_bytes(b[:4])

    def u64(self, b: memoryview) -> int:
        return self.from_bytes(b[:8])

    def to_bytes(self, i: int, length: int) -> bytes:
        return i.to_bytes(length, self.value)


@dataclass
class Go12State:
    functab: memoryview
    filetab: memoryview

    binary: ByteOrder = ByteOrder.LITTLE_ENDIAN

    quantum: int = 0
    ptrsize: int = 0

    nfunctab: int = 0
    nfiletab: int = 0

    file_map: Dict[str, int] = field(default_factory=dict)
    strings: Dict[int, str] = field(default_factory=dict)

    def uintptr(self, b: memoryview) -> int:
        return self.binary.from_bytes(b[:self.ptrsize])

    def string(self, data: memoryview, off: int) -> str:
        res = self.strings.get(off, None)
        if res is not None:
            return res

        # strlen
        size = 0
        for b in data[off:]:
            if b == 0:
                break
            size += 1

        res = data[off:off + size].tobytes().decode(SYMS_ENCODING)

        self.strings[off] = res
        return res

    # initFileMap initializes the map from file name to file number.
    def init_file_map(self) -> None:
        if self.file_map is not None:
            return

        self.file_map = {
            self.string(self.binary.u32(self.filetab[4 * i:])): i
            for i in range(1, self.nfiletab)
        }

    # go12MapFiles adds to m a key for every file in the Go 1.2 LineTable.
    # Every key maps to obj. That's not a very interesting map, but it provides
    # a way for callers to obtain the list of files in the program.
    def go12_map_files(self, m: Dict[str, Obj], obj: Obj) -> None:
        self.init_file_map()
        for file_name in self.file_map.keys():
            m[file_name] = obj

    def go12_funcs(self, line_table: "LineTable") -> List[Func]:
        data = line_table.data
        n = len(self.functab) // self.ptrsize // 2
        funcs = []
        for i in range(n):
            entry = self.uintptr(self.functab[2 * i * self.ptrsize:])
            end = self.uintptr(self.functab[(2 * i + 2) * self.ptrsize:])
            info = data[self.uintptr(self.functab[(2 * i + 1) * self.ptrsize:]):]
            frame_size = self.binary.u32(info[self.ptrsize + 2 * 4:])
            funcs.append(Func(
                sym=Sym(
                    value=entry,
                    type=ord("T"),
                    name=self.string(data, self.binary.u32(info[self.ptrsize:])),
                    go_type=0,
                ),
                entry=entry,
                end=end,
                line_table=line_table,
                frame_size=frame_size,
            ))
        return funcs


# NOTE(rsc): This is wrong for GOARCH=arm, which uses a quantum of 4,
# but we have no idea whether we're using arm or not. This only
# matters in the old (pre-Go 1.2) symbol table format, so it's not worth
# fixing.

OLD_QUANTUM = 1

GO12_MAGIC = 0xfffffffb


@dataclass
class LineTable:
    data: memoryview
    PC: int = 0
    line: int = 0

    checked_go12: bool = False
    _go12: Optional[Go12State] = None

    @property
    def go12(self) -> Optional[Go12State]:
        if not self.checked_go12:
            self.checked_go12 = True
            self.go12_init()
        return self._go12

    def go12_init(self) -> None:
        try:
            # Check header: 4-byte magic, two zeros, pc quantum, pointer size.
            if len(self.data) < 16 or self.data[4] != 0 or self.data[5] != 0 or \
               (self.data[6] != 1 and self.data[6] != 2 and self.data[6] != 4) or \
               (self.data[7] != 4 and self.data[7] != 8):  # pointer size
                logger.info("invalid go12 aux header")
                return

            if int.from_bytes(self.data[:4], "big") == GO12_MAGIC:
                binary = ByteOrder.BIG_ENDIAN
            elif int.from_bytes(self.data[:4], "little") == GO12_MAGIC:
                binary = ByteOrder.LITTLE_ENDIAN
            else:
                logger.info("invalid go12 magic")
                return

            quantum = self.data[6]
            ptrsize = self.data[7]
            nfunctab = binary.from_bytes(self.data[8:8 + ptrsize])
            functab = self.data[8 + ptrsize:]
            functabsize = (nfunctab * 2 + 1) * ptrsize
            fileoff = binary.from_bytes(functab[functabsize:functabsize + 4])
            functab = functab[:functabsize]
            filetab = self.data[fileoff:]
            nfiletab = binary.from_bytes(filetab[:4])
            filetab = filetab[:nfiletab * 4]

            self._go12 = Go12State(
                quantum=quantum,
                ptrsize=ptrsize,
                nfunctab=nfunctab,
                nfiletab=nfiletab,
                functab=functab,
                filetab=filetab,
            )
        # If we panic parsing, assume it's not a Go 1.2 symbol table.
        except Exception:
            raise

    def slice(self, pc: int) -> "LineTable":
        data, pc, line = self.parse(pc, -1)
        return LineTable(data=data, PC=pc, line=line)

    def parse(self, target_pc: int, target_line: int) -> Tuple[memoryview, int, int]:
        # The PC/line table can be thought of as a sequence of
        #  <pc update>* <line update>
        # batches. Each update batch results in a (pc, line) pair,
        # where line applies to every PC from pc up to but not
        # including the pc of the next pair.

        # Here we process each update individually, which simplifies
        # the code, but makes the corner cases more confusing.
        b, pc, line = self.data, self.PC, self.line
        while pc <= target_pc and line != target_line and len(b) > 0:
            code = b[0]
            b = b[1:]
            if code == 0:
                if len(b) < 4:
                    b = b[0:0]
                else:
                    val = int.from_bytes(b[:4], "big")
                    b = b[4:]
                    line += val
            elif code <= 64:
                line += code
            elif code <= 128:
                line -= code - 64
            else:
                pc += OLD_QUANTUM * (code - 128)
                continue
            pc += OLD_QUANTUM
        return b, pc, line


def new_line_table(data: bytes, text: int) -> LineTable:
    return LineTable(data=memoryview(data), PC=text, line=0)



import logging
from dataclasses import dataclass, field
from typing import Optional, List, Dict, Iterator
from .pclntab import LineTable, ByteOrder, Sym, Obj, Func, SYMS_ENCODING


# https://golang.org/src/debug/gosym/symtab.go


logger = logging.getLogger(__name__)


class DecodingError(Exception):
    pass


@dataclass
class Table:
    syms: List[Sym] = field(default_factory=list)
    funcs: List[Func] = field(default_factory=list)
    files: Dict[str, Obj] = field(default_factory=dict)
    objs: List[Obj] = field(default_factory=list)

    go12line: Optional[LineTable] = None


@dataclass
class sym:
    value: int = 0
    typ: int = 0
    name: bytes = b""
    gotype: int = 0


LITTLE_ENDIAN_SYMTAB     = bytes((0xFD, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00))
BIG_ENDIAN_SYMTAB        = bytes((0xFF, 0xFF, 0xFF, 0xFD, 0x00, 0x00, 0x00))
OLD_LITTLE_ENDIAN_SYMTAB = bytes((0xFE, 0xFF, 0xFF, 0xFF, 0x00, 0x00))


def walksymtab(data: memoryview) -> Iterator[sym]:
    if len(data) == 0:
        return

    new_table_format = False
    order = ByteOrder.BIG_ENDIAN
    data_magic = data[:7].tobytes()  # for .startswith
    if data_magic.startswith(OLD_LITTLE_ENDIAN_SYMTAB):
        data = data[6:]
        order = ByteOrder.LITTLE_ENDIAN
    elif data_magic.startswith(BIG_ENDIAN_SYMTAB):
        new_table_format = True
    elif data_magic.startswith(LITTLE_ENDIAN_SYMTAB):
        new_table_format = True
        order = ByteOrder.LITTLE_ENDIAN

    ptrsz = 0
    if new_table_format:
        if len(data) < 8:
            raise DecodingError("unexpected EOF")
        ptrsz = data[7]
        if ptrsz not in (4, 8):
            return DecodingError("invalid pointer size")
        data = data[8:]

    s = sym()
    p = data
    while len(p) >= 4:
        typ: int
        if new_table_format:
            # Symbol type, value, Go type.
            typ = p[0] & 0x3F
            wide_value = (p[0] & 0x40) != 0
            go_type = (p[0] & 0x80) != 0
            if typ < 26:
                typ += ord("A")
            else:
                typ += ord("a") - 26
            s.typ = typ
            p = p[1:]

            if wide_value:
                if len(p) < ptrsz:
                    raise DecodingError("unexpected EOF")

                s.value = order.from_bytes(p[0:ptrsz])
                p = p[ptrsz:]
            else:
                s.value = 0
                shift = 0
                while len(p) > 0 and p[0] & 0x80 != 0:
                    s.value |= p[0] & 0x7F << shift
                    shift += 7
                    p = p[1:]
                if len(p) == 0:
                    return DecodingError("unexpected EOF")

                s.value |= p[0] << shift
                p = p[1:]
            if go_type:
                if len(p) < ptrsz:
                    raise DecodingError("unexpected EOF")
                # fixed-width go type
                s.gotype = order.from_bytes(p[0:ptrsz])
                p = p[ptrsz:]
        else:
            # value, symbol type
            s.value = order.from_bytes(p[0:4])
            if len(p) < 5:
                raise DecodingError("unexpected EOF")
            typ = p[4]
            if typ & 0x80 == 0:
                raise DecodingError("bad symbol type")
            typ = typ & ~0x80
            s.typ = typ
            p = p[5:]

        # Name.
        i: int
        nnul: int
        for i in range(len(p)):
            if p[i] == 0:
                nnul = 1
                break

        if typ in b"Zz":
            p = p[i+nnul:]
            for i in range(0, len(p), 2):
                if p[i] == 0 and p[i + 1] == 0:
                    nnul = 2
                    break

        if len(p) < i + nnul:
            raise DecodingError("unexpected EOF")
        s.name = p[0:i].tobytes()
        i += nnul
        p = p[i:]

        if not new_table_format:
            if len(p) < 4:
                raise DecodingError("unexpected EOF")
            s.gotype = order.from_bytes(p[:4])
            p = p[4:]

        yield s


def new_table(symtab: memoryview, pcln: LineTable) -> Optional[Table]:
    t = Table()
    if pcln.go12 is not None:
        go12 = pcln.go12
        t.go12line = pcln

    fname: Dict[int, str] = {}
    nf = 0
    nz = 0
    lasttyp = 0

    for s in walksymtab(symtab):
        ts = Sym()
        t.syms.append(ts)
        ts.type = s.typ
        ts.value = s.value
        ts.go_type = s.gotype
        if s.typ in b"Zz":
            if lasttyp not in b"Zz":
                nz += 1
            for i in range(0, len(s.name), 2):
                elt_idx = int.from_bytes(s.name[i:i + 2], "big")
                elt = fname.get(elt_idx, None)
                if elt is None:
                    raise DecodingError("bad filename code")
                if ts.name and ts.name[-1] != "/":
                    ts.name += "/"
                ts.name += elt
        else:
            ts.name = s.name.decode(SYMS_ENCODING).replace("·", ".")

        if s.typ in b"f":
            fname[s.value] = ts.name
        elif s.typ in b"TtLl":
            nf += 1
        lasttyp = s.typ

    if t.go12line is not None:
        obj = Obj()
        t.objs = [obj]
        go12.go12_map_files(t.files, obj)
    else:
        t.objs = [Obj() for _ in range(nz)]

    lastf = 0
    i = 0
    while i < len(t.syms):
        sym = t.syms[i]
        # path symbol
        if sym.type in b"Zz":
            # Go 1.2 binaries have the file information elsewhere. Ignore.
            if t.go12line is None:
                i += 1
                continue

            # Finish the current object
            if obj is not None:
                obj.funcs = t.funcs[lastf:]
            lastf = len(t.funcs)

            # Start new object
            obj = Obj()
            t.objs.append(obj)

            # Count & copy path symbols
            for end in range(i + 1, len(t.syms)):
                c = t.syms[end].type
                if c not in b"Zz":
                    break
            obj.paths = t.syms[i:end]
            i = end
            continue
        elif sym.type in b"TtLl":
            if t.funcs:
                t.funcs[-1].end = sym.value
            if sym.name in {"runtime.etext", "etext"}:
                i += 1
                continue

            np = 0
            na = 0
            end = 0
            for end in range(i + 1, len(t.syms)):
                cur_sym_type = t.syms[end].type
                if cur_sym_type in b"TtLlZz":
                    break
                elif cur_sym_type in b"p":
                    np += 1
                elif cur_sym_type in b"a":
                    na += 1

            # Fill in the function symbol
            fn = Func()
            t.funcs.append(fn)
            sym.func = fn
            fn.sym = sym
            fn.entry = sym.value
            fn.obj = obj
            if t.go12line is not None:
                # All functions share the same line table.
                fn.line_table = t.go12line
            elif pcln is not None:
                fn.line_table = pcln.slice(fn.entry)
                pcln = fn.line_table

            for j in range(i, end):
                ls = t.syms[j]
                if ls.type in b"m":
                    fn.frame_size = s.value
                elif ls.type in b"p":
                    fn.params.append(ls)
                elif ls.type in b"a":
                    fn.locals.append(ls)
            i = end
            continue
        else:
            i += 1

    if t.go12line is not None and nf == 0:
        logger.info("reading go12 symbols")
        t.funcs = go12.go12_funcs(t.go12line)

    if obj is not None:
        obj.funcs = t.funcs[lastf:]

    return t
